#include "rtl865x/rtl_types.h"

#ifndef RTL865X_TEST
#ifdef __linux__
#include <linux/netdevice.h>
#include <linux/interrupt.h>
#include <linux/skbuff.h>
#endif
#include "swNic2.h"
#endif
#include "rtl865x/mbuf.h"
#include "rtl865x/assert.h"
#include "rtl865x/rtl8651_tblDrv.h"
#include "rtl865x/rtl8651_tblAsicDrv.h"

#ifdef CONFIG_RTL865X_PPP
#ifdef __linux__
#include <linux/timer.h>
#endif
#endif


#ifdef RTL865X_TEST
//------------------------  MODULE TEST CODE GLUE STARTS HERE---------------//
//This is not real Linux sk_buff. Simplified for module test.
struct sk_buff {
        /* These two members must be first. */
        struct sk_buff  * next;                 /* Next buffer in list                          */
        struct sk_buff  * prev;                 /* Previous buffer in list                      */
        struct sk_buff_head * list;             /* List we are on                               */
        struct sock     *sk;                    /* Socket we are owned by                       */
        unsigned int    len;                    /* Length of actual data                        */
        unsigned int    data_len;
        unsigned int    csum;                   /* Checksum                                     */
        unsigned char   cloned;                 /* head may be cloned (check refcnt to be sure). */
        unsigned int    truesize;               /* Buffer size                                  */

        unsigned char   *head;                  /* Head of buffer                               */
        unsigned char   *data;                  /* Data head pointer                            */
        unsigned char   *tail;                  /* Tail pointer                                 */
        unsigned char   *end;                   /* End pointer                                  */
};

struct sk_buff_head {
        struct sk_buff  * next;
        struct sk_buff  * prev;

        uint32          qlen;
};

struct skb_shared_info {
        unsigned int    nr_frags;
        struct sk_buff  *frag_list;
};

typedef struct { } spinlock_t;
int32 spin_lock_irqsave(spinlock_t *spinlock, int32 s){return 0;}
int32 spin_unlock_irqrestore(spinlock_t *spinlock, int32 s){return 0;}
int32 swNic_isrReclaim(uint32 rxDescIdx, struct rtl_pktHdr*pPkthdr,struct rtl_mBuf *pMbuf){}
int32 rtlglue_drvMutexLock(void){return 0;}
int32 rtlglue_drvMutexUnlock(void){return 0;}
int32 rtlglue_mbufMutexLock(void){return 0;}
int32 rtlglue_mbufMutexUnlock(void){return 0;}
void local_irq_restore(void){}
void local_irq_save(void){}

void rtlglue_getMacAddress(ether_addr_t * macAddress, uint32 * number){
	uint8 mac[] = { 0x00, 0x01, 0x02, 0x03, 0x04, 0x08 };
        memcpy((int8*)macAddress, (int8 *)mac, 6);
        *number = 8;
}

void *rtlglue_malloc(uint32 NBYTES){return malloc(NBYTES);}
void rtlglue_free(void *APTR){free(APTR);}
int32 rtlglue_mbufClusterRefcnt(void *buffer, uint32 id, uint32 *count, int8 operation){
        if (operation==0){
                if (count)
                        *count=1;
        }
        return SUCCESS;
}
int32 rtlglue_mbufAllocCluster(void **buffer, uint32 size, uint32 *id){
        struct sk_buff *skb;
	void* data;
	skb = (struct sk_buff *)malloc((sizeof(struct sk_buff_head) + size + sizeof(struct skb_shared_info)));
	data = skb + sizeof(struct sk_buff_head);
	skb->truesize = size + sizeof(struct sk_buff);
	skb->head = data;
	skb->data = data;
	skb->tail = data;
	skb->end = data+size;
	skb->len = 0;
        skb->cloned = 0;
        skb->data_len = 0;

	*id = (int)skb;
	*buffer = (void*)((uint32)skb->data|0xa0000000);
	return SUCCESS;
}
int32 rtlglue_mbufFreeCluster(void *buffer, uint32 size, uint32 id){
	free((void *)id);
        return 0;
}

void rtlglue_extDeviceSend(uint16 vid, int32 linkId, struct rtl_pktHdr *orig_pktHdr){
	freeMbuf(((struct rtl_pktHdr*)orig_pktHdr)->ph_mbuf);

}

 void rtlglue_drvSend(void * pkthdr){
	struct rtl_mBuf* new_m;
	// duplicate packet
	
	new_m = mBuf_dupPacket(((struct rtl_pktHdr*)pkthdr)->ph_mbuf, MBUF_DONTWAIT);	
	if (!new_m){
		
		printf("@mbuf duplicate error!\n");
		return;
	}

	// free original packet
	freeMbuf(((struct rtl_pktHdr*)pkthdr)->ph_mbuf);

	// enqueue new packet
	if (mbufList_enqueue(new_m)!=0){

		printf("@mbuf enqueue error!\n");
	}	
	return;
}
void * rtlglue_mbufClusterToData(void *buffer){
	return buffer;
}

void rtlglue_getRingSize(uint32 *rx, uint32 *tx)
{
	if (rx)
		*rx = 8;
	if (tx)
		*tx = 8;
}

int32 rtlglue_reclaimRxBD(uint32 rxDescIdx,struct rtl_pktHdr *pThisPkthdr, struct rtl_mBuf *pThisMbuf){pThisPkthdr->ph_rxdesc = -1; mBuf_driverFreeMbufChain(pThisMbuf);  return 0;};
int32 rtlglue_mbufTsleep(uint32 wait_channel){return 0;}
int32 rtlglue_mbufTwakeup(uint32 wait_channel){return 0;}



//------------------------MODULE  TEST CODE GLUE ENDS HERE---------------//





#else /*RTL865X_TEST*/
//------------------------  LINUX DEPEND GLUE STARTS HERE-------------//

#define RTLGLUE_MAX_DEVLINKS	10
//Don't use rtl8650DevLink[0]. It is reserved for FWD engine to send a bcast pkt to PCI/WLAN
//if rtlglue_extDeviceSend() is called with linkId=0, then this pkt should be bcast to all WDS links. 
struct DevLinkRecord {
	int8 devName[16];
	uint16 vid;
	uint32 portmask;
	struct net_device *pci_device;
	//void (*descFreeFunc)(void *);
} rtl8650DevLink[RTLGLUE_MAX_DEVLINKS]; 


#ifdef CONFIG_RTL865X_PPP
#define MAX_LINUX_TIMERS_USED	16
struct timer_list pppTimer[MAX_LINUX_TIMERS_USED];	/* Media monitoring timer. */
uint32 *timerId[MAX_LINUX_TIMERS_USED];
#endif

int32 rtlglue_init(void){
#ifdef CONFIG_RTL865X_PPP
	int32 i;
	struct timer_list *t;
	for(i=0;i<MAX_LINUX_TIMERS_USED;i++){
		timerId[i]=NULL;
		memset(&pppTimer[i], 0, sizeof(struct timer_list));
	}
#endif
	memset(&rtl8650DevLink[0],0,sizeof(struct DevLinkRecord));
        return 0;
}

/* rtlglue_registerTimerMs:  OS dependent timer registration function.
*	input: 
*		eventId: a value-result parameter,  for saving registered timer event  id. 
*		eventSerialNumber: a value-result parameter, for saving the unique serial number returned with event id.
*		callback_func: the registering callback function for execution
*		callback_func_arg:  arguments for callback_func
*		msec: timeout in units of milli-seconds.
*
*	return:
*		SUCCESS: if timer event successfully registered.
*		FAILED: if timer event registration failed.
*
*	Porintg comments: This glue function helps register a timer event to the ported OS.
*		If the native timer service provided by the ported OS is a kernel service(which means atomic operation),
*		the return value of 2nd parameter should always be zero. If the timer service is provided by a
*		"soft" timer thread, then the return value of 'eventSerialNumber' must be unique each time this glue 
*		function is called. The 'callback_func' shall be coded to check the first parameter during called back to prevent
*		erroneous operation between two racing threads.
*
*/
int32 rtlglue_registerTimerMs(uint32 *eventId,uint32 *eventSerialNumber,int32 (* callback_func)(uint32,void *), void *callback_func_arg, uint32 msec){

#ifdef CONFIG_RTL865X_PPP
	int32 i, empty=-1;
	struct timer_list *t;
	if(!eventId)
		return FAILED;
	for(i=0;i<MAX_LINUX_TIMERS_USED;i++){
		if(eventId==timerId[i]){
			//already registered. Reuse this timer object.
			t=&pppTimer[empty];
			mod_timer(t, jiffies + (msec*HZ)/1000);
			goto out;
		}
		if(!timerId[i])
			empty=i; //last usable entry seen
	}
	if(empty==-1){
		//all timer objects exhausted.
		return FAILED;
	}

	//use this empty timer object to register a new timer event to Linux kernel
	t=&pppTimer[empty];
	init_timer(t);
	t->expires = jiffies + (msec*HZ)/1000;//system tick to expire
	t->data = (unsigned long)callback_func_arg;
	t->function = callback_func;
	add_timer(t);
out:	
	*eventSerialNumber=0;
	*eventId=t;
#endif
	return SUCCESS;
}


/* rtlglue_cancelTimer:  OS dependent timer registration function.
*	input: 
*		eventId: The timer event  id to be cancelled.
*		eventSerialNumber: The serial number of timer event id.
*	return:
*		SUCCESS: Timer event successfully cancelled.
*		FAILED: Failed to cancel timer event
*/

int32 rtlglue_cancelTimer(uint32 eventId,uint32 eventSerialNumber)
{
#ifdef CONFIG_RTL865X_PPP
	int32 i;
	struct timer_list *t;
	if(!eventId)
		return FAILED;
	for(i=0;i<MAX_LINUX_TIMERS_USED;i++){
		if(eventId==timerId[i]){
			//timer found. delete this timer from timer list
			del_timer(&pppTimer[i]);
			goto out;
		}
	}
	return FAILED;
out:	
#endif
        return SUCCESS;
}
static spinlock_t mbuf_lock = SPIN_LOCK_UNLOCKED;
static int __mbuf_flags;
int dd_cc;

int32 rtlglue_mbufMutexLock(void){
	if (dd_cc>=1){
		dd_cc++;
		return 0;
	}	
	spin_lock_irqsave(&mbuf_lock,__mbuf_flags);
	//enableLbcTimeout();
	dd_cc++;
        return 0;
}

int32 rtlglue_mbufMutexUnlock(void){
	dd_cc--;
	if (!dd_cc)
		spin_unlock_irqrestore(&mbuf_lock,__mbuf_flags);
        return 0;
}

int32 rtlglue_mbufTsleep(uint32 wait_channel){
        return 0;
}

int32 rtlglue_mbufTwakeup(uint32 wait_channel){
        return 0;
}

int32 rtlglue_mbufClusterRefcnt(void *buffer, uint32 id, uint32 *count, int8 operation){
	if (operation==0)
	{
		if (count)
			*count=1;

	}
        return SUCCESS;
}

void * rtlglue_mbufClusterToData(void *buffer){
	
	return (buffer);
}

int32 rtlglue_mbufAllocCluster(void **buffer, uint32 size, uint32 *id){
	struct sk_buff *skb;
	skb =	dev_alloc_skb (size);
	if (skb)
	{
		*id = (int)skb;
#ifdef CONFIG_RTL865X_CACHED_NETWORK_IO
		*buffer = (void*)((u32)skb->data);
#else
		*buffer = (void*)((u32)skb->data|0xa0000000);
#endif
       		 return SUCCESS;
	}
	else 
	{
		printk("skb exhausted. alloc failed\n");
		return FAILED;
	}
}
#define DBG 1
int32 rtlglue_mbufFreeCluster(void *buffer, uint32 size, uint32 id){
	struct sk_buff *skb=(struct sk_buff*)id;

#if DBG
	if (!skb)
	{
		printk(" %s %d skb null data %x\n",__FILE__,__LINE__,(uint32)buffer);
		return 0;
	}			
#endif
	dev_kfree_skb_any(skb);
        return 0;
}
static spinlock_t int_lock = SPIN_LOCK_UNLOCKED;
int int_cc=0;
static int __flags;
int32 rtlglue_drvMutexLock(void){
	
	if (int_cc>=1) 
	{
		int_cc++;
		return 0;
	}
	spin_lock_irqsave(&int_lock,__flags);
	//enableLbcTimeout();	
	int_cc++;
        return 0;
}

int32 rtlglue_drvMutexUnlock(void){
	int_cc--;
	if (!int_cc)
	spin_unlock_irqrestore(&int_lock,__flags);
        return 0;
}




void rtlglue_drvSend(void *data)
{
	struct rtl_pktHdr *pktHdr=(struct rtl_pktHdr *)data;
	//printk("rtlglue_drvSend:");
	if(FAILED==swNic_write((void *)pktHdr) &&((pktHdr->ph_rxdesc<0)||(pktHdr->ph_rxdesc>=totalRxPkthdr))){
		//buffer was allocated by upper layer. Free both mbuf and skbuf here...
		mBuf_freeMbufChain(pktHdr->ph_mbuf);
	}
}


int32 rtlglue_reclaimRxBD(uint32 rxDescIdx,struct rtl_pktHdr *pThisPkthdr, struct rtl_mBuf *pThisMbuf){
	int32 s,retval;
	spin_lock_irqsave(rtl865xSpinlock,s);
	assert(!pThisMbuf->m_next);
	//reclaim the pkthdr--mbuf--cluster...
	retval=swNic_isrReclaim(rxDescIdx, pThisPkthdr,  pThisMbuf);
	spin_unlock_irqrestore(rtl865xSpinlock,s);
	return retval;
}

#ifdef CONFIG_RTL865X_MULTILAYER_BSP
//This function register a WLAN driver instance or a WDS link to a locally maintained database. 
//This function is not pat of 8651 driver glue interface but is implemented here to demonstrate how 
//to create linkId and bind it with OS-dependent driver interface.
//Valid linkId number can be any value except 0. linkId=0 would be used by 8651 driver when 
//rtlglue_extDeviceSend() is called to tell WLAN driver that a broadcast must be sent to all WLAN 
//cards or all WDS links on the designated vlan..
// ***This function is not called/used by 8651 driver 
int32 devglue_regExtDevice(int8 *devName, uint16 vid, uint8 extPortNum, uint32 *linkId){//, uint16 vid){//, void (*descFreeFunc)(void *)){
	int8 chipVersion[16];
	int32 i;
	if(!devName){//||!descFreeFunc){
		printk("%s:No device name given\n", __FUNCTION__);
		return FAILED;
	}
	if(!vid||vid>4094){
		printk("%s:Last parameter can't be NULL\n", __FUNCTION__);
		return FAILED;
	}

	rtl8651_getChipVersion(chipVersion, sizeof(chipVersion), NULL);
	if(chipVersion[strlen(chipVersion)]=='B'){
		printk("%s:WLAN acceleration requires RTL865xB chip\n", __FUNCTION__);
		return FAILED;
	}else if(!extPortNum||extPortNum>RTL8651_PORT_NUMBER+rtl8651_totalExtPortNum)
	{
		printk("%s:Illegal 865XB ext port 0x%x\n", __FUNCTION__, extPortNum);
		return FAILED;
	}
	if(!linkId){
		printk("%s:Last parameter can't be NULL\n", __FUNCTION__);
		return FAILED;
	}

	//start from 1. 0 is reserved for "ALL WDS ports"
	for(i=1;i<RTLGLUE_MAX_DEVLINKS;i++)
		if(rtl8650DevLink[i].devName[0]=='\0')
			break; //found a unused extension port
	if(i>=RTLGLUE_MAX_DEVLINKS){
		printk("No more free ext port\n");
		return FAILED;
	}
	rtl8650DevLink[i].pci_device=dev_get_by_name(devName);
	if(!rtl8650DevLink[i].pci_device){
		printk("No such device:%s\n", devName);
		return FAILED;
	}
	strcpy(rtl8650DevLink[i].devName, devName);
	rtl8650DevLink[i].vid=vid;
	rtl8650DevLink[i].portmask=1<<extPortNum;
	if(FAILED==rtl8651_addPeripheralExtPort(extPortNum)){
		printk("Failed when register peripheral extport %d\n", extPortNum);
		return FAILED;
	}
	*linkId = i;//i is the registered extension port number. User may use this number from now on
	return SUCCESS; 
}


int32 devglue_unregExtDevice(uint32 linkId){
	int32 i;
	uint32 portmask;
	if(RTLGLUE_MAX_DEVLINKS<linkId)
		return FAILED;
	portmask=rtl8650DevLink[linkId].portmask;
	for(i=0;portmask==1;i++)
		portmask>>=1;		
	if(FAILED==rtl8651_delPeripheralExtPort(i)){
		printk("Failed when deregister peripheral extport %d\n", i);
		return FAILED;
	}
	rtl8650DevLink[linkId].devName[0]='\0';
	rtl8650DevLink[linkId].vid=rtl8650DevLink[linkId].portmask=0;
	rtl8650DevLink[linkId].pci_device=NULL;	
	return SUCCESS;
}


//This function is for OS-dependent part of WLAN driver to get 'devName's liniId number 
//before sending the pkt to 8651 river. Each WLAN pkt to be fast forwarded by 8651 ASIC 
//must be passed with its belonging linkId number to 8651 driver when 
//rtl8651_fwdEngineExtPortRecv() is called.
// ***This function is not called/used by 8651 driver 
int32 devglue_getLinkIDByName(int8 *devName){
	int32 i;
	//start from 1. 
	//0 is reserved for "ALL WDS ports"	
	for(i=1;i<RTLGLUE_MAX_DEVLINKS;i++){
		if(strcmp(devName, rtl8650DevLink[i].devName)==0)
			return i;
	}
	return FAILED;
}

EXPORT_SYMBOL(devglue_regExtDevice);
EXPORT_SYMBOL(devglue_getLinkIDByName);
EXPORT_SYMBOL(devglue_unregExtDevice);

int chkValidAddr(uint32 addr){
	if(addr&0x80000000){
		if(addr>>28==0x8 || addr>>28==0xa||addr>>28==0xb)
			return 0;
	}
	printk("%08x invalid!!\n", addr);
	return 1;
}
int checkSkbAddr(struct sk_buff *skb){
	if(chkValidAddr((uint32)skb->head))
		return 1;
	if(chkValidAddr((uint32)skb->data))
		return 2;
	if(chkValidAddr((uint32)skb->tail))
		return 3;
	if(chkValidAddr((uint32)skb->end))
		return 4;
	return 0;
}

#define PKT_BUF_SZ 2048
static  struct sk_buff *re865x_mbuf2skb(struct rtl_pktHdr* pPkt) {
	struct sk_buff *new_skb, *pkt_skb;
	struct rtl_mBuf *pkt_mbuf=pPkt->ph_mbuf;
	struct sk_buff *retval=NULL;
	assert(pPkt->ph_mbuf);	
	
	#ifdef RTL865X_DEBUG
	if(wlan_acc_debug>1){
		uint8 *data= pkt_mbuf->m_data;
		uint8 macStr[sizeof"ff:ff:ff:ff:ff:ff"];			
		ether_ntoa_r((ether_addr_t *)&data[6], macStr);
		printk("\nPCI Tx(%d)%s->",pPkt->ph_rxdesc,macStr);
		ether_ntoa_r((ether_addr_t *)&data[0], macStr);
		printk("%s  0x%x  Len:%d...\n",macStr,*((uint16 *)(data+12)), pPkt->ph_len);
	}
	#endif
	if((pPkt->ph_rxdesc<0)||(pPkt->ph_rxdesc>=totalRxPkthdr)){
		//cfliu: <0: from upper layer stack, >=totalRxPkthdr: duplicated from fwdEngine
		struct rtl_mBuf *m= pkt_mbuf;
		assert(m);
		pkt_skb = (struct sk_buff *)pkt_mbuf->m_extClusterId;
		if(pkt_skb->len==0){
			pkt_skb->data+=mBuf_leadingSpace(pkt_mbuf);
			skb_put(pkt_skb, pPkt->ph_len);
		}
		//free mbuf/pkthdr
		mBuf_freeOneMbufPkthdr(m, (void * *) &m->m_extbuf, &m->m_extClusterId,&m->m_extsize);
		return pkt_skb;
	}
	//not from upper layer stack, allocate a new pkt buffer and skb header.
	//new pkt buffer would be linked with pPkt so as to refill swnic's rx ring
	//new skb would have its skb->data points to pPkt->ph_mbuf->m_data.
	new_skb = dev_alloc_skb (PKT_BUF_SZ);
	if (new_skb) 
	{
		int32 align,s;
		pkt_skb = (struct sk_buff *)pkt_mbuf->m_extClusterId;
		pkt_skb->ip_summed = CHECKSUM_NONE;
		//force set length to 0
		pkt_skb->data = pkt_skb->tail = pkt_skb->head ;
		pkt_skb->len=0;
		//adjust data pointer by mbuf
		skb_reserve(pkt_skb, 16+mBuf_leadingSpace(pkt_mbuf));
		//set real pkt length
		skb_put(pkt_skb, pPkt->ph_len);
#ifdef CONFIG_RTL865X_MBUF_HEADROOM
		align=CONFIG_RTL865X_MBUF_HEADROOM;
#endif
		//refill Rx ring, use new buffer
		mBuf_attachCluster(pkt_mbuf, new_skb->data, (uint32)new_skb, 2048, 0, align);
		spin_lock_irqsave(rtl865xSpinlock,s);
		if (swNic_isrReclaim(pPkt->ph_rxdesc, pPkt, pkt_mbuf))
			goto free;
		spin_unlock_irqrestore(rtl865xSpinlock,s);
		retval= pkt_skb; 
	}else{
free:
		mBuf_freeMbufChain(pkt_mbuf);
		retval= NULL;
	}
	return retval;
}

//For more info about this function, read API definition in rtl_glue.h
void rtlglue_extDeviceSend(uint16 vid, int32 linkId, struct rtl_pktHdr *orig_pktHdr){
	struct sk_buff *skb;
	struct rtl_pktHdr *thisPkthdr=orig_pktHdr, *firstCopy=NULL;
	int32 ret, thisLink=linkId;
	struct rtl_mBuf *newMbuf;

	//linkId>0: unicast
	//linkId=0: bcast from LAN port, send to all existing links.
	//linkId<0: bcast from abs(linkId) WLAN card or WDS link. Would avoid sending to that link again.

	if(linkId>0){
		goto direct_unicast;
	}		
	//linkId<=0, it's a bcast pkt or unknown unicast
	thisLink=1;//start from 1
	//duplicate and free original packet ASAP to avoid rx ring runout.
	newMbuf= mBuf_dupPacket(orig_pktHdr->ph_mbuf, MBUF_DONTWAIT);
	mBuf_driverFreeMbufChain(orig_pktHdr->ph_mbuf);
	if(!newMbuf){
		#ifdef RTL865X_DEBUG
		if(wlan_acc_debug>2)
			printk("Bcast pkt duplicated failed!!!\n");
		#endif
		return;		
	}
	
	#ifdef RTL865X_DEBUG
	if(wlan_acc_debug>2) {
		printk("Bcast pkt duplicated. New:%x\n", (uint32)newMbuf);
	}
	#endif
	firstCopy=newMbuf->m_pkthdr;//firstCopy would be copied to each extport later...

next_link:
	//last pkt sent freed. prepare a copy for this link from firstCopy
	assert(firstCopy);
	newMbuf= mBuf_dupPacket(firstCopy->ph_mbuf, MBUF_DONTWAIT);
	if(!newMbuf){
		goto end_bcast;
	}
next_link_pktready:
	//send this copied pkt to thisLink.
	#ifdef RTL865X_DEBUG
	if(wlan_acc_debug>2) {
		printk("dup pkt for bcast to link %d\n", thisLink);
	}
	#endif
	thisPkthdr=newMbuf->m_pkthdr;

	assert((thisPkthdr->ph_rxdesc<0)||(thisPkthdr->ph_rxdesc>=totalRxPkthdr));
	if (rtl8650DevLink[thisLink].vid!=vid||
		thisLink==-linkId||!rtl8650DevLink[thisLink].pci_device||
		!netif_running(rtl8650DevLink[thisLink].pci_device)
	){
		#ifdef RTL865X_DEBUG
		if(wlan_acc_debug>1) 
			printk("bcast to link %d aborted.Freed!\n", thisLink);
		#endif

		//loop until next registered link found in this vlan or quit.
		do{
			if(thisLink>=RTLGLUE_MAX_DEVLINKS){
				mBuf_driverFreeMbufChain(thisPkthdr->ph_mbuf);
				goto end_bcast;
			}
			thisLink++;
			if(rtl8650DevLink[thisLink].devName[0]&&rtl8650DevLink[thisLink].vid==vid){
				#ifdef RTL865X_DEBUG
				if(wlan_acc_debug>1) 
					printk("Try send to next link %d\n", thisLink);
				#endif
				goto next_link_pktready;
			}
		}while(1);
		assert(0);//should not come here.
	}
	skb= (struct sk_buff *)re865x_mbuf2skb(thisPkthdr);
	//mbuf/pkthdr is already freed after re865x_mbuf2skb()
	if(!skb)
		goto end_bcast;
	skb->dev=rtl8650DevLink[thisLink].pci_device;

	#ifdef RTL865X_DEBUG
	if(wlan_acc_debug>1) {
		printk("to pci (linkId %d) %s...\n", thisLink, skb->dev->name);
	}
	#endif
	ret=dev_queue_xmit(skb);
	if(ret<0){
		#ifdef RTL865X_DEBUG
		if(wlan_acc_debug>1)
			printk("dev_queue_xmit dropped. Reason: %d \n", ret);
		#endif
	}

	if(thisLink<RTLGLUE_MAX_DEVLINKS){
		thisLink++;
		goto next_link;
	}

end_bcast:
	//remember to free the template 'firstCopy' before we return
	mBuf_driverFreeMbufChain(firstCopy->ph_mbuf);
	return;


direct_unicast:
	#ifdef RTL865X_DEBUG
	if((orig_pktHdr->ph_mbuf->m_data[0]&1)==1){
		printk("XXX Error: Bcast pkt send to linkId %d directly!!!!\n", linkId);
		mBuf_driverFreeMbufChain(orig_pktHdr->ph_mbuf);
		return;
	}
	#endif

	if ((vid!=rtl8650DevLink[thisLink].vid)||!rtl8650DevLink[thisLink].pci_device||!netif_running(rtl8650DevLink[thisLink].pci_device)){
		#ifdef RTL865X_DEBUG
		if(wlan_acc_debug>1) 
			printk("link %d Not running. (%d,%d) Freed!\n", thisLink, vid,rtl8650DevLink[thisLink].vid);
		#endif
		//free unicast pkt since port not ready for send.
		assert(thisPkthdr==orig_pktHdr);
		mBuf_driverFreeMbufChain(thisPkthdr->ph_mbuf);
		return;
	}

	skb= (struct sk_buff *)re865x_mbuf2skb(thisPkthdr);
	//mbuf/pkthdr is already freed after re865x_mbuf2skb()
	if(!skb)
		return;
	skb->dev=rtl8650DevLink[thisLink].pci_device;

	#ifdef RTL865X_DEBUG
	if(wlan_acc_debug>1) {
		printk("to pci (linkId %d) %s...\n", thisLink, skb->dev->name);
	}
	#endif
	ret=dev_queue_xmit(skb);
	if(ret<0){
		#ifdef RTL865X_DEBUG
		if(wlan_acc_debug>1)
			printk("dev_queue_xmit dropped. Reason: %d \n", ret);
		#endif
	}
	return;	
}

#endif


#define BDINFO_ADDR 0xbfc04000
void rtlglue_getMacAddress(ether_addr_t * macAddress, uint32 * number) {
	rtl8651_memcpy((char*)macAddress,(char*)BDINFO_ADDR,6);
}

void rtlglue_getRingSize(uint32 *rx, uint32 *tx){
	if(rx)
		*rx = totalRxPkthdr;
	if(tx)
		*tx = totalTxPkthdr;
}

//extern int jiffies;
/*
 * Note: the unit of jiffies is 10ms.
 */

/*
 * getsectime() returns seconds.
 */
uint32 getsectime(void) {
	return jiffies/100;
		
}

/*
 * getmstime() returns milli-second (1/1000 sec).
 */
uint32 getmstime(void) {
        return jiffies*10;
}

uint32 rtlglue_getmstime( uint32* pTime )
{
	return *pTime = getmstime();
}



static int totalmem;
void *rtlglue_malloc(size_t NBYTES) {
	totalmem = totalmem+NBYTES;
	return kmalloc(NBYTES,GFP_KERNEL);
}

void rtlglue_free(void *APTR) {
	kfree(APTR);
}
//------------------------  LINUX DEPEND GLUE STARTS HERE-------------//
#endif /*RTL865X_TEST*/
